# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

import warnings

import jax

if not jax.config.jax_enable_x64:
    raise RuntimeError(f"jax_enable_x64 must be set to True to use cuQuantum Python JAX")

from .operator_action import operator_action
from .pysrc import (
    ElementaryOperator,
    MatrixOperator,
    OperatorTerm,
    Operator
)



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
cuDensityMat context classes.
"""

import atexit
import logging
from typing import List

import jax

from cuquantum.bindings import cudensitymat as cudm
from nvmath.internal import typemaps

from .operator import Operator


class CudensitymatContext:
    """
    cuDensityMat library context.

    This class holds the library handle and the workspace descriptor handle, which are used for all
    operator actions. A specific operator action context is created for each operator,
    appended to _contexts, and retrieved when the specific operator action is invoked.
    """

    _handle = None
    _workspace_desc = None
    _operator_ptrs = set()  # Store only pointers, not operator objects (to avoid leaking tracers)
    _contexts = {}

    logger = logging.getLogger("cudensitymat-jax.CudensitymatContext")

    @classmethod
    def _maybe_create_handle_and_workspace(cls):
        """
        Create the handle and workspace for the context if they are not already created.
        """
        if cls._handle is None:
            cls.logger.info(f"Initializing CudensitymatContext")
            cls._handle = cudm.create()
            cls.logger.info(f"Created handle at {hex(cls._handle)}")

            if cls._workspace_desc is None:
                cls._workspace_desc = cudm.create_workspace(cls._handle)
                cls.logger.info(f"Created workspace descriptor at {hex(cls._workspace_desc)}")
            else:  # handle is None but workspace is not None
                raise RuntimeError("Workspace descriptor and handle should be created at the same time")
        else:
            if cls._workspace_desc is None:  # handle is not None but workspace is None
                raise RuntimeError("Workspace descriptor and handle should be created at the same time")

    @classmethod
    def maybe_create_context(cls, op: Operator, device: jax.Device, batch_size: int, purity: cudm.StatePurity) -> None:
        """
        Create the OperatorActionContext for a new operator.
        """
        cls._maybe_create_handle_and_workspace()

        # Check if there is already a context for the operator and if not, create it.
        # NOTE: We check if a context exists using the operator's pointer. The pointer is created
        # during context initialization, so we check if it's already been set and has a context.
        # We only store the operator pointer, not the operator object itself, to avoid
        # leaking JAX tracers when the operator is used inside transformations like jax.vjp.
        # _contexts is indexed by opaque pointers since the operator objects themselves change
        # when they are flattened and unflattened during PyTree manipulation.
        if op._ptr is None or op._ptr not in cls._contexts:
            ctx = OperatorActionContext(op, device, batch_size, purity)
            cls._contexts[op._ptr] = ctx
            cls._operator_ptrs.add(op._ptr)

            cls.logger.info(f"Created OperatorActionContext for operator {hex(id(op))}")

    @classmethod
    def get_context(cls, op: Operator) -> "OperatorActionContext":
        """
        Get the OperatorActionContext for a given operator.
        """
        # TODO: Raise an error with error message when op._ptr is not in cls._contexts.
        return cls._contexts[op._ptr]

    @classmethod
    def free(cls):
        """
        Free opaque handles to the library.
        
        Note: We don't store operator objects in contexts to avoid leaking JAX tracers.
        Operator objects and their GPU handles will be cleaned up by Python's garbage
        collector when they're no longer referenced. Here we only clean up the state
        handles and workspace that are managed by contexts.
        """
        cls.logger.info(f"Freeing CudensitymatContext")

        # Free all state handles from contexts
        for ctx in cls._contexts.values():
            ctx.free()

        # Free workspace and library handle
        if cls._workspace_desc is not None:
            cudm.destroy_workspace(cls._workspace_desc)
            cls._workspace_desc = None

        if cls._handle is not None:
            cudm.destroy(cls._handle)
            cls._handle = None

        # Clear tracking dictionaries
        CudensitymatContext._operator_ptrs.clear()
        CudensitymatContext._contexts.clear()


atexit.register(CudensitymatContext.free)


class OperatorActionContext:
    """
    Operator action context.
    """

    logger = logging.getLogger("cudensitymat-jax.OperatorActionContext")

    def __init__(self,
                 op: Operator,
                 device: jax.Device,
                 batch_size: int,
                 purity: cudm.StatePurity
                 ) -> None:
        """
        Initialize OperatorActionContext.

        Args:
            op: The operator object for operator action.
            device: The device to use in operator action. If None, the first GPU device will be used.
            batch_size: Batch size of the operator action.
            purity: Purity of the state.
        """
        self.logger.info(f"Initializing OperatorActionContext")

        # Instance attributes.
        self.device = device
        self.batch_size = batch_size
        self.state_purity = purity

        # Derived attributes from operator.
        self._space_mode_extents = op.dims
        self._data_type = typemaps.NAME_TO_DATA_TYPE[op.dtype.name]
        self._compute_type = typemaps.NAME_TO_COMPUTE_TYPE[op.dtype.name]

        # A unified buffer size for all primitives.
        self._required_buffer_size = 0

        # Create opaque handle to the operator.
        op._create(CudensitymatContext._handle)
        self._operator = op._ptr
        # Note: We don't store the operator object to avoid leaking JAX tracers

        # Create opaque handles to the input and output states.
        self._state_in = cudm.create_state(
            CudensitymatContext._handle,
            self.state_purity,
            len(self._space_mode_extents),
            self._space_mode_extents,
            self.batch_size,
            self._data_type
        )
        self.logger.debug(f"Created input state at {hex(self._state_in)}")

        self._state_out = cudm.create_state(
            CudensitymatContext._handle,
            self.state_purity,
            len(self._space_mode_extents),
            self._space_mode_extents,
            self.batch_size,
            self._data_type
        )
        self.logger.debug(f"Created output state at {hex(self._state_out)}")

        # The state adjoints are to be set in create_adjoint_buffers when backward rule is called.
        self._state_in_adj = None
        self._state_out_adj = None

        # Attributes to be assigned in operator_action.
        self.base_op_ptrs: List[int] | None = None
        self.is_elem_op: List[int] | None = None
        self.num_state_components: int | None = None
        self.op_term_batched_coeffs_tmp: List[int] | None = None
        self.op_prod_batched_coeffs_tmp: List[int] | None = None

    def create_adjoint_buffers(self):
        """
        Create adjoint buffers for the input and output states.
        """
        # Create opaque handles to the input and output state adjoints.
        self._state_in_adj = cudm.create_state(
            CudensitymatContext._handle,
            self.state_purity,
            len(self._space_mode_extents),
            self._space_mode_extents,
            self.batch_size,
            self._data_type
        )
        self.logger.debug(f"Created input state adjoint at {hex(self._state_in_adj)}")

        self._state_out_adj = cudm.create_state(
            CudensitymatContext._handle,
            self.state_purity,
            len(self._space_mode_extents),
            self._space_mode_extents,
            self.batch_size,
            self._data_type
        )
        self.logger.debug(f"Created output state adjoint at {hex(self._state_out_adj)}")

    def free(self):
        """
        Free opaque handles to the library.
        """
        self.logger.info(f"Freeing OperatorActionContext")

        if self._state_out_adj is not None:
            cudm.destroy_state(self._state_out_adj)
            self.logger.debug(f"Destroyed output state adjoint at {hex(self._state_out_adj)}")
            self._state_out_adj = None

        if self._state_in_adj is not None:
            cudm.destroy_state(self._state_in_adj)
            self.logger.debug(f"Destroyed input state adjoint at {hex(self._state_in_adj)}")
            self._state_in_adj = None

        if self._state_out is not None:
            cudm.destroy_state(self._state_out)
            self.logger.debug(f"Destroyed output state at {hex(self._state_out)}")
            self._state_out = None

        if self._state_in is not None:
            cudm.destroy_state(self._state_in)
            self.logger.debug(f"Destroyed input state at {hex(self._state_in)}")
            self._state_in = None



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Utility functions for cuQuantum Python JAX.
"""

from jax.interpreters.batching import BatchTracer

from cuquantum.bindings import cudensitymat as cudm


def maybe_expand_dim(bufs, ndim):
    """
    Expand to a leading batch dimension when jax.vmap is applied or when the state is non-batched.
    """
    if isinstance(bufs[0], BatchTracer) or len(bufs[0].shape) in [ndim, 2 * ndim]:
        bufs = tuple([buf.reshape((1, *buf.shape)) for buf in bufs])
    return bufs


def maybe_squeeze_dim(bufs, ndim):
    """
    Squeeze the leading batch dimension when jax.vmap is applied or when the state is non-batched.
    """
    if isinstance(bufs[0], BatchTracer) or (
        len(bufs[0].shape) in [ndim + 1, 2 * ndim + 1] and bufs[0].shape[0] == 1
    ):
        bufs = tuple([buf.reshape(*buf.shape[1:]) for buf in bufs])
    return bufs


def get_state_batch_size_and_purity(state_in_bufs, ndim, default_batch_size):
    """
    Get the batch size and purity of the state.
    """
    # First check all state buffer shapes.
    shape = state_in_bufs[0].shape
    for buf in state_in_bufs[1:]:
        if buf.shape != shape:
            raise ValueError("All input state buffers must have the same shape.")
    if len(shape) not in [ndim, 2 * ndim, ndim + 1, 2 * ndim + 1]:
        raise ValueError("The dimensions of the input state do not match the dimensions of the operator.")

    # Extract batch size based on the four values ndim, 2 * ndim, ndim + 1, 2 * ndim + 1.
    if isinstance(state_in_bufs[0], BatchTracer):
        batch_size = default_batch_size
    else:
        if len(shape) % ndim == 0:
            batch_size = 1
        else:  # ndim + 1 or 2 * ndim + 1
            batch_size = shape[0]

    # Extract purity based on the four values ndim, 2 * ndim, ndim + 1, 2 * ndim + 1.
    if len(shape) // ndim == 1:
        purity = cudm.StatePurity.PURE
    else:  # 2 * ndim or 2 * ndim + 1
        purity = cudm.StatePurity.MIXED

    return batch_size, purity


def check_and_return_final_batch_size(state_in_bufs, state_batch_size, op_batch_size, default_batch_size):
    """
    Check and return the final batch size.
    """
    if isinstance(state_in_bufs[0], BatchTracer):
        return default_batch_size
    else:
        if len({1, state_batch_size, op_batch_size}) > 2:  # the set should be either {1} or {1, N}
            raise ValueError("The batch size of the input state does not match the batch size of the operator.")
        return max(state_batch_size, op_batch_size)



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Matrix operator class in cuDensityMat.
"""

import logging
from typing import Tuple

import jax
import jax.numpy as jnp

from cuquantum.bindings import cudensitymat as cudm
from nvmath.internal import typemaps


@jax.tree_util.register_pytree_node_class
class MatrixOperator:
    """
    PyTree class for cuDensityMat's matrix operator.
    """

    logger = logging.getLogger("cudensitymat-jax.MatrixOperator")

    def __init__(self,
                 data: jax.Array | jax.ShapeDtypeStruct,
                 callback: cudm.WrappedTensorCallback | None = None,
                 grad_callback: cudm.WrappedTensorGradientCallback | None = None
                 ) -> None:
        """
        Initialize a MatrixOperator object.

        Args:
            data: Data specification of the matrix operator. If ``callback`` is ``None``, ``data``
                should be a ``jax.Array``; otherwise, ``data`` should be a ``jax.ShapeDtypeStruct``.
            callback: Forward callback for the matrix operator.
            grad_callback: Gradient callback for the matrix operator.
        """
        if type(data) is object:
            # TODO: data becomes object() during AD tracing. tree_unflatten should be written properly
            # to handle this case.
            pass
        elif callback is not None and not isinstance(data, jax.ShapeDtypeStruct):
            raise RuntimeError("data must be a jax.ShapeDtypeStruct when the data buffer "
                               "is dynamically constructed from a callback.")

        if isinstance(data, (jax.Array, jax.ShapeDtypeStruct)):

            # Set data and batch size.
            if data.ndim % 2 == 0:
                # Expanding to a leading dimension 1 is necessary since we are taking 0 as the batch
                # dimension when passing to ffi_lowering.
                if isinstance(data, jax.Array):
                    self.data = jnp.expand_dims(data, 0)
                    self._to_be_allocated = False
                else:
                    self.data = jnp.zeros((1, *data.shape), dtype=data.dtype)
                    self._to_be_allocated = True
                self.batch_size = 1
            else:  # batched
                if isinstance(data, jax.Array):
                    self.data = data
                    self._to_be_allocated = False
                else:
                    self.data = jnp.zeros(data.shape, dtype=data.dtype)
                    self._to_be_allocated = True
                self.batch_size = data.shape[0]

            # Set other attributes derived from data.
            self.num_modes: int = len(self.data.shape) // 2
            if self.data.shape[-2 * self.num_modes:-self.num_modes] != self.data.shape[-self.num_modes:]:
                raise ValueError("Data must have the same shape on the bra and ket modes.")
            self.mode_extents: Tuple[int, ...] = self.data.shape[-self.num_modes:]
            self.dtype: jnp.dtype = self.data.dtype

        elif type(data) is object:  # data is object() during AD tracing.
            # Dummy variables for derived attributes.
            self.data: object = data
            self.batch_size: int = 1
            self.num_modes: int = 0
            self.mode_extents: Tuple[int, ...] = ()
            self.dtype: jnp.dtype = jnp.dtype(float)
            self._to_be_allocated: bool = False
        else:
            # We don't put object() in the error message (like data can also be object())
            # because that case is in internal JAX implementation.
            raise TypeError("data must be a jax.Array or jax.ShapeDtypeStruct.")

        # Callbacks.
        self.callback: cudm.WrappedTensorCallback | None = callback
        self.grad_callback: cudm.WrappedTensorGradientCallback | None = grad_callback

        self._ptr: int | None = None
        self._is_elementary: int | None = None

    def copy(self) -> "MatrixOperator":
        """
        Copy the matrix operator.
        """
        if self._to_be_allocated and type(self.data) is not object:
            data = jax.ShapeDtypeStruct(self.data.shape, self.data.dtype)
        else:
            data = jnp.copy(self.data)

        mat_op = MatrixOperator(data, self.callback, self.grad_callback)
        mat_op._ptr = self._ptr
        mat_op._is_elementary = self._is_elementary
        mat_op._to_be_allocated = self._to_be_allocated
        return mat_op

    def tree_flatten(self):
        """
        Flatten the matrix operator PyTree.
        """
        children = (self.data, self._ptr, self._is_elementary)
        aux_data = (self.callback, self.grad_callback, self._to_be_allocated)
        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """
        Unflatten the matrix operator PyTree.
        """
        data, _ptr, _is_elementary = children
        callback, grad_callback, _to_be_allocated = aux_data

        if _to_be_allocated and type(data) is not object:
            data = jax.ShapeDtypeStruct(data.shape, data.dtype)

        inst = cls(data, callback, grad_callback)
        inst._ptr = _ptr
        inst._is_elementary = _is_elementary
        inst._to_be_allocated = _to_be_allocated
        return inst

    def _create(self, handle):
        """
        Create opaque handle to the matrix operator.
        """
        # NOTE: This is to catch the gradient attachment case where grad_callback is provided
        # but callback is not.
        if self.callback is None and self.grad_callback is not None:
            def f(t, args, storage):
                pass
            callback = cudm.WrappedTensorCallback(f, cudm.CallbackDevice.GPU)
        else:
            callback = self.callback

        # Create opaque handle to the matrix operator.
        if self._ptr is None:
            if self.batch_size == 1:
                self._ptr = cudm.create_matrix_operator_dense_local(
                    handle,
                    self.num_modes,
                    self.mode_extents,
                    typemaps.NAME_TO_DATA_TYPE[self.dtype.name],
                    0,  # buffer pointer to be attached in the XLA layer
                    callback,
                    self.grad_callback,
                )
            else:
                self._ptr = cudm.create_matrix_operator_dense_local_batch(
                    handle,
                    self.num_modes,
                    self.mode_extents,
                    self.batch_size,
                    typemaps.NAME_TO_DATA_TYPE[self.dtype.name],
                    0,  # buffer pointer to be attached in the XLA layer
                    callback,
                    self.grad_callback,
                )

            self.logger.debug(f"Created matrix operator at {hex(self._ptr)}")

        # Set in _create to prevent tracing.
        if self._is_elementary is None:
            self._is_elementary = 0

    def _destroy(self):
        """
        Destroy opaque handle to the matrix operator.
        """
        if self._ptr is not None:
            cudm.destroy_matrix_operator(self._ptr)
            self.logger.debug(f"Destroyed matrix operator at {hex(self._ptr)}")
            self._ptr = None



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Operator term class in cuDensityMat.
"""

import ctypes
import logging
from typing import List, Tuple, Sequence, Type

import jax
import jax.numpy as jnp

from cuquantum.bindings import cudensitymat as cudm

from .elementary_operator import ElementaryOperator
from .matrix_operator import MatrixOperator


@jax.tree_util.register_pytree_node_class
class OperatorTerm:
    """
    PyTree class for cuDensityMat's operator term.
    """

    logger = logging.getLogger("cudensitymat-jax.OperatorTerm")

    def __init__(self, dims: Sequence[int]) -> None:
        """
        Initialize an OperatorTerm object.

        Args:
            dims: Hilbert space dimensions.
        """
        # Input argument.
        self.dims: Tuple[int, ...] = tuple(dims)

        # Attributes for handling arguments in append.
        self.op_prods: List[Tuple[ElementaryOperator | MatrixOperator, ...]] = []
        self.modes: List[Tuple[int, ...]] = []
        self.conjs: List[Tuple[bool, ...]] = []
        self.duals: List[Tuple[bool, ...]] = []

        self.coeffs: List[float | complex] = []
        self.static_coeffs: List[jax.Array | None] = []
        self.total_coeffs: List[jax.ShapeDtypeStruct | None] = []
        self.coeff_callbacks: List[cudm.WrappedScalarCallback | None] = []
        self.coeff_grad_callbacks: List[cudm.WrappedScalarGradientCallback | None] = []

        self.batch_size: int = 1
        self.dtype: jnp.dtype | None = None

        # Temporary pointers to be replaced by real pointers at execution time.
        self.static_coeffs_ptrs: List[int] = []
        self.total_coeffs_ptrs: List[int] = []
        self._static_coeffs_ptr_objs: List[ctypes.c_short] = []
        self._total_coeffs_ptr_objs: List[ctypes.c_short] = []

        self._batch_sizes: List[int] = []
        self._op_prod_types: List[Type[ElementaryOperator] | Type[MatrixOperator]] = []
        self._ptr: int | None = None

    def copy(self) -> "OperatorTerm":
        """
        Copy the operator term.
        """
        op_term = OperatorTerm(self.dims)
        for (
            op_prod,
            modes,
            conjs,
            duals,
            coeff,
            static_coeffs,
            total_coeffs,
            coeff_callback,
            coeff_grad_callbacks
        ) in zip(
            self.op_prods,
            self.modes,
            self.conjs,
            self.duals,
            self.coeffs,
            self.static_coeffs,
            self.total_coeffs,
            self.coeff_callbacks,
            self.coeff_grad_callbacks
        ):
            # This is due to empty tuple appended to preserve length.
            if conjs == ():
                conjs = None
            
            # This is due to modes set to all Hilbert space modes internally
            # for reference implementation.
            if isinstance(op_prod[0], MatrixOperator):
                modes = None

            op_prod_ = tuple(base_op.copy() for base_op in op_prod)
            static_coeffs_ = jnp.copy(static_coeffs) if static_coeffs is not None else None
            total_coeffs_ = total_coeffs if total_coeffs is not None else None
            op_term.append(op_prod_,
                           modes=modes,
                           conjs=conjs,
                           duals=duals,
                           coeff=coeff,
                           static_coeffs=static_coeffs_,
                           total_coeffs=total_coeffs_,
                           coeff_callback=coeff_callback,
                           coeff_grad_callback=coeff_grad_callbacks)
        
        # Preserve pointer information to maintain PyTree structure consistency.
        # This is critical for custom VJP: the backward output must have the same
        # PyTree structure as the forward input, including pointer values.
        op_term.static_coeffs_ptrs = self.static_coeffs_ptrs.copy()
        op_term.total_coeffs_ptrs = self.total_coeffs_ptrs.copy()
        op_term._ptr = self._ptr
        
        return op_term

    def tree_flatten(self):
        """
        Flatten the operator term into a PyTree.
        """
        children = (self.op_prods, self.static_coeffs, self.static_coeffs_ptrs)
        aux_data = (
            self.dims,
            self.batch_size,
            self.modes,
            self.conjs,
            self.duals,
            self.coeffs,
            self.total_coeffs,
            self.total_coeffs_ptrs,
            self.coeff_callbacks,
            self.coeff_grad_callbacks,
            self.dtype,
            self._batch_sizes,
            self._op_prod_types,
            self._ptr
        )
        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """
        Unflatten the operator term from a PyTree.
        """
        op_prods, static_coeffs, static_coeffs_ptrs = children
        (
            dims,
            batch_size,
            modes,
            conjs,
            duals,
            coeffs,
            total_coeffs,
            total_coeffs_ptrs,
            coeff_callbacks,
            coeff_grad_callbacks,
            dtype,
            _batch_sizes,
            _op_prod_types,
            _ptr
        ) = aux_data

        # Create the instance.
        inst = cls(dims)

        # Populate traced attributes.
        inst.op_prods = op_prods
        inst.static_coeffs = static_coeffs
        inst.static_coeffs_ptrs = static_coeffs_ptrs
        inst.total_coeffs = total_coeffs
        inst.total_coeffs_ptrs = total_coeffs_ptrs

        # Populate static attributes.
        inst.batch_size = batch_size
        inst.modes = modes
        inst.conjs = conjs
        inst.duals = duals
        inst.coeffs = coeffs
        inst.coeff_callbacks = coeff_callbacks
        inst.coeff_grad_callbacks = coeff_grad_callbacks
        inst.dtype = dtype
        inst._batch_sizes = _batch_sizes
        inst._op_prod_types = _op_prod_types
        inst._ptr = _ptr
        return inst

    def _check_dtype(self, op_prod):
        """
        Check if all elementary or matrix operators have the same data type.
        """
        if self.dtype is None:
            # If the data type is not set, set it to the data type of the first operator.
            self.dtype = op_prod[0].dtype
            for op in op_prod[1:]:
                if op.dtype != self.dtype:
                    raise ValueError("All elementary or matrix operators must have the same data type.")
        else:
            # If the data type is set, check if all elementary or matrix operators
            # have the same data type.
            for op in op_prod:
                if op.dtype != self.dtype:
                    raise ValueError("All elementary or matrix operators must have the same data type.")

    def _check_and_append_op_prod_type(self, op_prod):
        """
        Check if all terms in an operator product are of the same type.
        """
        op_prod_type = type(op_prod[0])
        for op in op_prod[1:]:
            if not isinstance(op, op_prod_type):
                raise ValueError("All terms in an operator product must be of the same type.")
        self._op_prod_types.append(op_prod_type)

    def _check_and_append_batch_size(self,
                                     op_prod: Sequence[ElementaryOperator | MatrixOperator],
                                     static_coeffs: jax.Array | None,
                                     total_coeffs: jax.Array | None
                                     ) -> None:
        """
        Check if all terms in an operator product have batch size 1 or N.
        """
        # Extract the batch size for the operator product.
        op_prod_batch_size = max([base_op.batch_size for base_op in op_prod])
        for base_op in op_prod:
            if base_op.batch_size not in (1, op_prod_batch_size):
                raise ValueError("All basic operators in an operator product must have batch size 1 or N.")

        # Extract the batch sizes of the static and total coefficients.
        if static_coeffs is not None and total_coeffs is None:
            coeffs_batch_size = static_coeffs.shape[0]
        elif static_coeffs is None and total_coeffs is not None:
            coeffs_batch_size = total_coeffs.shape[0]
        elif static_coeffs is not None and total_coeffs is not None:
            if static_coeffs.shape[0] != total_coeffs.shape[0]:
                raise ValueError("Static and total coefficients must have the same batch size if both are provided.")
            coeffs_batch_size = static_coeffs.shape[0]
        else:
            coeffs_batch_size = 1
        self._batch_sizes.append(coeffs_batch_size)

        batch_size = max(op_prod_batch_size, coeffs_batch_size)
        if self.batch_size == 1:
            self.batch_size = batch_size
        else:
            if batch_size not in (1, self.batch_size):
                raise ValueError("Batch size in this operator product does not match batch size of this operator term.")

    def append(self,
               op_prod: Sequence[ElementaryOperator | MatrixOperator],
               *,
               modes: Sequence[int] | None = None,
               conjs: Sequence[bool] | None = None,
               duals: Sequence[bool] | None = None,
               coeff: float | complex = 1.0,
               static_coeffs: jax.Array | None = None,
               total_coeffs: jax.ShapeDtypeStruct | None = None,
               coeff_callback: cudm.WrappedScalarCallback | None = None,
               coeff_grad_callback: cudm.WrappedScalarGradientCallback | None = None
               ) -> None:
        """
        Append an elementary or matrix product to an operator term.

        Args:
            op_prod: Product of elementary or matrix operators to be appended.
            modes: Modes acted on by the operator product.
            conjs: Conjugations in the operator product. Only applies to MatrixOperators.
            duals: Dualities of the operator product.
            coeff: Coefficient of the operator product.
            static_coeffs: Static batched coefficients of the operator product.
            total_coeffs: Total batched coefficients of the operator product.
            coeff_callback: Forward callback for the coefficients.
            coeff_grad_callback: Gradient callback for the coefficients.
        """
        # Only require total_coeffs when using batched coefficients (static_coeffs provided)
        if coeff_callback is not None and static_coeffs is not None:
            if not isinstance(total_coeffs, jax.ShapeDtypeStruct):
                raise RuntimeError("For dynamically constructed batched coefficients, "
                                   "total_coeffs must be a jax.ShapeDtypeStruct.")

        # Check if all elementary or matrix operators have the same data type.
        self._check_dtype(op_prod)
        self._check_and_append_op_prod_type(op_prod)
        self._check_and_append_batch_size(op_prod, static_coeffs, total_coeffs)

        # Check consistency and append modes, conjs and duals.
        if self._op_prod_types[-1] is ElementaryOperator:
            # Modes have to specified for elementary operators.
            if modes is None:
                raise ValueError("Modes acted on must be specified for elementary operators.")

            # Check all modes are in Hilbert space.
            if not set(modes) <= set(range(len(self.dims))):
                raise ValueError("Modes acted on must be in the Hilbert space, i.e. between 0 and len(self.dims) - 1")

            # Check length of modes acted on are the same as combined number of modes in the operator product.
            if len(modes) != sum([elem_op.num_modes for elem_op in op_prod]):
                raise ValueError(f"Number of modes acted on {len(modes)} does not match combined number of modes in the operator product.")

            # Check mode extents of each elementary operator match corresponding qubit dimensions.
            modes_index = 0
            for elem_op in op_prod:
                if elem_op.mode_extents != tuple(
                    [self.dims[modes[i]] for i in range(modes_index, modes_index + elem_op.num_modes)]
                ):
                    raise ValueError("Mode extents of each elementary operator must match corresponding qubit dimensions.")
                modes_index += elem_op.num_modes

            # Check that matrix conjugations cannot be specified for elementary operators.
            if conjs is not None:
                raise ValueError("Matrix conjugations cannot be specified for elementary operators.")

            # Check that number of duals matches number of modes.
            if duals is None:
                duals = (False,) * len(modes)
            else:
                if len(duals) != len(modes):
                    raise ValueError("Number of duals must match number of modes acted on for elementary operator product.")

            # For elementary operator product, we only need modes and duals.
            self.modes.append(tuple(modes))
            self.conjs.append(())  # empty tuple is appended here to preserve length
            self.duals.append(tuple(duals))

        else:  # matrix operator product
            # Check that mode extents match Hilbert space dimensions.
            for matrix_op in op_prod:
                if matrix_op.mode_extents != self.dims:
                    raise ValueError("Mode extents must match Hilbert space dimensions for matrix operators.")

            # Check that modes acted on cannot be specified for matrix operators.
            if modes is not None:
                raise ValueError("Modes acted on cannot be specified for matrix operators.")

            # Check consistency of conjs.
            if conjs is None:
                conjs = (False,) * len(op_prod)
            else:
                if len(conjs) != len(op_prod):
                    raise ValueError("Number of matrix conjugations must match number of operator products.")

            # Check that number of duals matches number of matrix operators.
            if duals is None:
                duals = (False,) * len(op_prod)
            else:
                if len(duals) != len(op_prod):
                    raise ValueError("Number of duals must match number of matrix operators.")

            # For matrix operator product, we only need conjs and duals.
            self.modes.append(tuple(range(len(self.dims))))  # used in reference implementation during testing
            self.conjs.append(tuple(conjs))
            self.duals.append(tuple(duals))

        # Populate instance attributes.
        self.op_prods.append(tuple(op_prod))
        self.coeffs.append(coeff)
        self.static_coeffs.append(static_coeffs)
        self.total_coeffs.append(total_coeffs)
        self.static_coeffs_ptrs.append(None)
        self.total_coeffs_ptrs.append(None)
        self.coeff_callbacks.append(coeff_callback)
        self.coeff_grad_callbacks.append(coeff_grad_callback)

    def __getitem__(self, index: int) -> Tuple[ElementaryOperator | MatrixOperator, ...]:
        """
        Get an operator product from the operator term.
        """
        return self.op_prods[index]

    def _create(self, handle):
        """
        Create opaque handle to the operator term.
        """
        # Create opaque handle to dependent elementary or matrix operators.
        for op_prod in self.op_prods:
            for elem_op in op_prod:
                elem_op._create(handle)

        # Create opaque handle to the operator term.    
        if self._ptr is None:
            self._ptr = cudm.create_operator_term(
                handle,
                len(self.dims),
                self.dims
            )
            self.logger.debug(f"Created operator term at {hex(self._ptr)}")

            for i in range(len(self.op_prods)):
                if self._op_prod_types[i] is ElementaryOperator:
                    if self._batch_sizes[i] == 1:
                        cudm.operator_term_append_elementary_product(
                            handle,
                            self._ptr,
                            len(self.op_prods[i]),
                            [elem_op._ptr for elem_op in self.op_prods[i]],
                            self.modes[i],
                            self.duals[i],
                            self.coeffs[i],
                            self.coeff_callbacks[i],
                            self.coeff_grad_callbacks[i]
                        )
                    else:  # batched
                        if self.static_coeffs[i] is not None:
                            # Create and store object for temporary static coefficients.
                            static_coeffs_ptr_obj = ctypes.c_short()
                            self._static_coeffs_ptr_objs.append(static_coeffs_ptr_obj)

                            # Extract pointer of the object for temporary static coefficients.
                            static_coeffs_ptr = ctypes.addressof(static_coeffs_ptr_obj) + 1
                            assert static_coeffs_ptr % 2 == 1, "Temporary static coefficients pointer must be an odd number."
                            self.static_coeffs_ptrs[i] = static_coeffs_ptr
                        else:
                            static_coeffs_ptr = 0

                        if self.total_coeffs[i] is not None:
                            # Create and store object for temporary pointer.
                            total_coeffs_ptr_obj = ctypes.c_short()
                            self._total_coeffs_ptr_objs.append(total_coeffs_ptr_obj)

                            # Extract pointer of the object for temporary total coefficients.
                            total_coeffs_ptr = ctypes.addressof(total_coeffs_ptr_obj) + 1
                            assert total_coeffs_ptr % 2 == 1, "Temporary total coefficients pointer must be an odd number."
                            self.total_coeffs_ptrs[i] = total_coeffs_ptr
                        else:
                            total_coeffs_ptr = 0

                        cudm.operator_term_append_elementary_product_batch(
                            handle,
                            self._ptr,
                            len(self.op_prods[i]),
                            [elem_op._ptr for elem_op in self.op_prods[i]],
                            self.modes[i],
                            self.duals[i],
                            self._batch_sizes[i],
                            static_coeffs_ptr,
                            total_coeffs_ptr,
                            self.coeff_callbacks[i],
                            self.coeff_grad_callbacks[i],
                        )
                else:  # MatrixOperator
                    if self._batch_sizes[i] == 1:
                        cudm.operator_term_append_matrix_product(
                            handle,
                            self._ptr,
                            len(self.op_prods[i]),
                            [mat_op._ptr for mat_op in self.op_prods[i]],
                            self.conjs[i],
                            self.duals[i],
                            self.coeffs[i],
                            self.coeff_callbacks[i],
                            self.coeff_grad_callbacks[i]
                        )
                    else:  # batched
                        if self.static_coeffs[i] is not None:
                            # Create and store object for temporary static coefficients.
                            static_coeffs_ptr_obj = ctypes.c_short()
                            self._static_coeffs_ptr_objs.append(static_coeffs_ptr_obj)

                            # Extract pointer of the object for temporary static coefficients.
                            static_coeffs_ptr = ctypes.addressof(static_coeffs_ptr_obj) + 1
                            assert static_coeffs_ptr % 2 == 1, "Temporary static coefficients pointer must be an odd number."
                            self.static_coeffs_ptrs[i] = static_coeffs_ptr
                        else:
                            static_coeffs_ptr = 0

                        if self.total_coeffs[i] is not None:
                            # Create and store object for temporary total coefficients.
                            total_coeffs_ptr_obj = ctypes.c_short()
                            self._total_coeffs_ptr_objs.append(total_coeffs_ptr_obj)

                            # Extract pointer of the object for temporary total coefficients.
                            total_coeffs_ptr = ctypes.addressof(total_coeffs_ptr_obj) + 1
                            assert total_coeffs_ptr % 2 == 1, "Temporary total coefficients pointer must be an odd number."
                            self.total_coeffs_ptrs[i] = total_coeffs_ptr
                        else:
                            total_coeffs_ptr = 0

                        cudm.operator_term_append_matrix_product_batch(
                            handle,
                            self._ptr,
                            len(self.op_prods[i]),
                            [mat_op._ptr for mat_op in self.op_prods[i]],
                            self.conjs[i],
                            self.duals[i],
                            self._batch_sizes[i],
                            static_coeffs_ptr,
                            total_coeffs_ptr,
                            self.coeff_callbacks[i],
                            self.coeff_grad_callbacks[i],
                        )

            self.logger.debug(f"Appended operator products to operator term at {hex(self._ptr)}")

    def _destroy(self):
        """
        Destroy opaque handle to the operator term.
        """
        self._static_coeffs_ptr_objs.clear()
        self._total_coeffs_ptr_objs.clear()

        # Destroy opaque handle to the operator term.
        if self._ptr is not None:
            cudm.destroy_operator_term(self._ptr)
            self.logger.debug(f"Destroyed operator term at {hex(self._ptr)}")
            self._ptr = None



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

import logging
from typing import Sequence, List, Tuple

import jax
import jax.numpy as jnp

from .pysrc.context import CudensitymatContext
from .pysrc.operator import Operator
from .pysrc.operator_action_prim import (
    operator_action_prim,
    operator_action_backward_diff_prim,
    OperatorActionPrimitive,
    OperatorActionBackwardDiffPrimitive
)
from .utils import (
    maybe_expand_dim,
    maybe_squeeze_dim,
    get_state_batch_size_and_purity,
    check_and_return_final_batch_size,
)

logger = logging.getLogger("cudensitymat-jax.operator_action")


def operator_action(op: Operator,
                    t: float,
                    state_in_bufs: jax.Array | Sequence[jax.Array],
                    params: jax.Array | None = None,
                    device: jax.Device | None = None,
                    batch_size: int = 1,
                    options: dict = {}
                    ) -> jax.Array | List[jax.Array]:
    """
    Compute the action of an operator on a state.

    Args:
        op: Operator to compute the action of.
        t: The time to compute the operator action at.
        state_in_bufs: Buffers of the input state components.
        params: Callback parameters used to construct the operator.
        device: Device to use for the operator action.
        batch_size: Batch size of the operator action.
        options: Dictionary of options for the operator action. Currently, the available options are names of the gradients attached to the callbacks. See ``example8_gradient_attachment.py`` for an example.

            - ``"op_term_grad_attr"``: Name of the operator term coefficient gradient attached to the coefficient gradient callback. Default is ``"scalar_grad"``.
            - ``"op_prod_grad_attr"``: Name of the operator product coefficient gradient attached to the coefficient gradient callback. Default is ``"scalar_grad"``.
            - ``"base_op_grad_attr"``: Name of the tensor gradient attached to the tensor gradient callback. Default is ``"tensor_grad"``.

    Returns:
        Buffers of the output state components.
    """
    logger.info(f"Calling operator_action")

    # Check if GPU is available and set to GPU 0.
    if device is None:
        gpu_devices = jax.devices('gpu')
        if gpu_devices == []:
            raise RuntimeError("No GPU devices found.")
        device = gpu_devices[0]
        logger.info("Running a single-GPU version of the library and setting device to GPU 0.")

    # Process input arguments.
    if isinstance(state_in_bufs, jax.Array):
        state_in_bufs = (state_in_bufs,)
    else:
        state_in_bufs = tuple(state_in_bufs)

    # Check state shape and maybe expand to a leading batch dimension.
    state_batch_size, purity = get_state_batch_size_and_purity(state_in_bufs, len(op.dims), batch_size)
    batch_size = check_and_return_final_batch_size(state_in_bufs, state_batch_size, op.batch_size, batch_size)

    state_in_bufs = maybe_expand_dim(state_in_bufs, len(op.dims))

    # Prepare library context for forward operator action.
    CudensitymatContext.maybe_create_context(op, device, batch_size, purity)
    
    # Store only the pointer to avoid leaking JAX tracers during transformations
    OperatorActionPrimitive.operator_ptr = op._ptr

    # Obtain the operator action context and assign corresponding attributes.
    op_act_ctx = CudensitymatContext.get_context(op)
    op_act_ctx.num_state_components = len(state_in_bufs)

    # Create lists for other_out attributes.
    other_out_types = []
    other_out_ptrs = []
    other_out_shape_dtypes = []

    other_in_types = []
    other_in_ptrs = []
    
    op_term_coeffs_indices = []
    op_prod_coeffs_indices = []
    base_op_indices = []

    # Extract temporary batched coefficient buffers for operator terms.
    count = 0
    for ptr, buf in zip(op.static_coeffs_ptrs, op.static_coeffs):
        if buf is not None and ptr is not None and ptr not in other_in_ptrs:
            other_in_types.append(4)
            other_in_ptrs.append(ptr)
            op_term_coeffs_indices.append(count)
        count += 1
    for ptr, buf in zip(op.total_coeffs_ptrs, op.total_coeffs):
        if buf is not None and ptr is not None and ptr not in other_out_ptrs:
            other_out_types.append(4)
            other_out_ptrs.append(ptr)
            other_out_shape_dtypes.append(buf)

    # Extract temporary batched coefficient buffers for operator products.
    count = 0
    static_coeffs = [x for op_term in op.op_terms for x in op_term.static_coeffs]
    total_coeffs = [x for op_term in op.op_terms for x in op_term.total_coeffs]
    static_coeffs_ptrs = [x for op_term in op.op_terms for x in op_term.static_coeffs_ptrs]
    total_coeffs_ptrs = [x for op_term in op.op_terms for x in op_term.total_coeffs_ptrs]
    for ptr, buf in zip(static_coeffs_ptrs, static_coeffs):
        if buf is not None and ptr is not None and ptr not in other_in_ptrs:
            other_in_types.append(3)
            other_in_ptrs.append(ptr)
            op_prod_coeffs_indices.append(count)
        count += 1
    for ptr, buf in zip(total_coeffs_ptrs, total_coeffs):
        if buf is not None and ptr is not None and ptr not in other_out_ptrs:
            other_out_types.append(3)
            other_out_ptrs.append(ptr)
            other_out_shape_dtypes.append(buf)

    # Assign certain static attributes to the operator action context.
    base_ops = [base_op for op_term in op.op_terms for op_prod in op_term.op_prods for base_op in op_prod]
    count = 0
    for base_op in base_ops:
        # Use callback being None or not to detect whether data is user-provided or to be allocated.
        if not base_op._to_be_allocated and base_op._ptr is not None:
            other_in_types.append(2 - base_op._is_elementary)  # 1 -> 1, 0 -> 2
            other_in_ptrs.append(base_op._ptr)
            base_op_indices.append(count)
        elif base_op._to_be_allocated and base_op._ptr is not None:
            other_out_types.append(2 - base_op._is_elementary)  # 1 -> 1, 0 -> 2
            other_out_ptrs.append(base_op._ptr)
            shape_dtype = jax.ShapeDtypeStruct(base_op.data.shape, base_op.data.dtype)
            other_out_shape_dtypes.append(shape_dtype)
        count += 1

    # Assign other output attributes to the operator action context.
    op_act_ctx.other_out_shape_dtypes = other_out_shape_dtypes
    op_act_ctx.other_out_types = other_out_types
    op_act_ctx.other_out_ptrs = other_out_ptrs

    op_act_ctx.other_in_types = other_in_types
    op_act_ctx.other_in_ptrs = other_in_ptrs

    op_act_ctx.op_term_coeffs_indices = op_term_coeffs_indices
    op_act_ctx.op_prod_coeffs_indices = op_prod_coeffs_indices
    op_act_ctx.base_op_indices = base_op_indices

    if params is None or params.shape == (0,):
        # Empty params causes a problem when reconstructing params from pointer 
        # in the bindings. We're providing a value here to guard against that.
        params = jnp.array([[0.0]])
    elif params.ndim == 1:
        params = jnp.array([params])
    else:
        if params.dtype != jnp.float64:
            raise ValueError("params must be a float64 array")
        
    op_act_ctx.options = options

    # Invoke operator action.
    state_out_bufs = _operator_action(op, t, state_in_bufs, params)
    state_out_bufs = maybe_squeeze_dim(state_out_bufs, len(op.dims))

    # Process output argument.
    if len(state_out_bufs) == 1:
        state_out_bufs = state_out_bufs[0]

    return state_out_bufs


@jax.custom_vjp
def _operator_action(op: Operator,
                     t: float,
                     state_in_bufs: Tuple[jax.Array, ...],
                     params: jax.Array,
                     ) -> List[jax.Array]:
    """
    Custom VJP rule for operator_action.
    """
    logger.info(f"Calling _operator_action")
    state_out_bufs, _ = _operator_action_fwd(op, t, state_in_bufs, params)
    return state_out_bufs


def _operator_action_fwd(op: Operator,
                         t: float,
                         state_in_bufs: Tuple[jax.Array, ...],
                         params: jax.Array,
                         ) -> Tuple[List[jax.Array], tuple]:
    """
    Forward rule for operator_action.
    """
    logger.info(f"Calling _operator_action_fwd")
    state_out_bufs = operator_action_prim(op, t, state_in_bufs, params)
    return state_out_bufs, (op, t, state_in_bufs, params)


def _operator_action_bwd(res: tuple, state_out_adj_bufs: jax.Array | Sequence[jax.Array]) -> tuple:
    """
    Backward rule for operator_action.

    Args:
        state_out_adj_bufs: Data buffers of the output state adjoint.
    """
    logger.info(f"Calling _operator_action_bwd")

    op, t, state_in_bufs, params = res
    
    # Prepare library context for backward operator action
    # Store only the pointer to avoid leaking JAX tracers during transformations
    OperatorActionBackwardDiffPrimitive.operator_ptr = op._ptr
    op_act_ctx = CudensitymatContext.get_context(op)
    op_act_ctx.create_adjoint_buffers()

    # Process input argument.
    if isinstance(state_out_adj_bufs, jax.Array):
        state_out_adj_bufs = (state_out_adj_bufs,)
    else:
        state_out_adj_bufs = tuple(state_out_adj_bufs)

    if len(state_in_bufs) != len(state_out_adj_bufs):
        raise ValueError("state_in_bufs and state_out_adj_bufs must have the same number of components.")

    params_grad, *state_in_adj_bufs = operator_action_backward_diff_prim(
        op, t, state_in_bufs, state_out_adj_bufs, params)
    
    if len(state_in_adj_bufs) == 1:
        state_in_adj_bufs = state_in_adj_bufs[0]
    
    # Lists used to store pointers to the different quantities, so that we know which gradient
    # has been set already.
    op_term_ptrs = []
    op_prod_ptrs = []
    base_op_ptrs = []

    options = op_act_ctx.options
    op_term_grad_attr = options.get("op_term_grad_attr", "scalar_grad")
    op_prod_grad_attr = options.get("op_prod_grad_attr", "scalar_grad")
    base_op_grad_attr = options.get("base_op_grad_attr", "tensor_grad")

    op_grad = op.copy()
    for i, (op_term, coeff_grad_callback) in enumerate(zip(op_grad.op_terms, op_grad.coeff_grad_callbacks)):
        # We need to set gradient for operator term coefficients only when there is coefficient
        # gradient callback, the coefficient gradient callback has a scalar_grad field, and gradient
        # on this operator term has not been set.
        set_op_term_coeffs_grad = (
            coeff_grad_callback is not None and
            hasattr(coeff_grad_callback.callback, op_term_grad_attr) and
            op_term._ptr not in op_term_ptrs
        )

        if op.static_coeffs[i] is None:  # non-batched
            if set_op_term_coeffs_grad:
                op_grad.coeffs[i] = getattr(coeff_grad_callback.callback, op_term_grad_attr)
                op_term_ptrs.append(op_term._ptr)
            # else:
            #     op_grad.coeffs[i] = 0.0j
        else:  # batched
            if set_op_term_coeffs_grad:
                op_grad.static_coeffs[i] = getattr(coeff_grad_callback.callback, op_term_grad_attr)
                op_term_ptrs.append(op_term._ptr)
            else:
                op_grad.static_coeffs[i] = jnp.zeros(op_grad.static_coeffs[i].shape, dtype=op_grad.static_coeffs[i].dtype)

        for j, (op_prod, coeff_grad_callback_) in enumerate(zip(op_term.op_prods, op_term.coeff_grad_callbacks)):
            # We need to set gradient for operator product coefficients only when there is coefficient
            # gradient callback, the coefficient gradient callback has a scalar_grad field, and gradient
            # on this operator product has not been set.
            op_prod_ptr = tuple([base_op._ptr for base_op in op_prod])
            set_op_prod_coeffs_grad = (
                coeff_grad_callback_ is not None and
                hasattr(coeff_grad_callback_.callback, op_prod_grad_attr) and
                op_prod_ptr not in op_prod_ptrs
            )

            if op_term.static_coeffs[j] is None:  # non-batched
                if set_op_prod_coeffs_grad:
                    op_grad.coeffs[j] = getattr(coeff_grad_callback_.callback, op_prod_grad_attr)
                    op_prod_ptrs.append(op_prod_ptr)
                # else:
                #     op_grad.coeffs[j] = 0.0j
            else:  # batched
                if set_op_prod_coeffs_grad:
                    op_prod_ptrs.append(op_prod_ptr)
                    op_term.static_coeffs[j] = getattr(coeff_grad_callback_.callback, op_prod_grad_attr)
                else:
                    op_term.static_coeffs[j] = jnp.zeros(
                        op_term.static_coeffs[j].shape,
                        dtype=op_term.static_coeffs[j].dtype
                    )

            for base_op in op_prod:
                # We need to set gradient for base operator data only when there is tensor
                # gradient callback, the tensor gradient callback has a tensor_grad field,
                # and gradient on this base operator has not been set.
                set_base_op_data_grad = (
                    base_op.grad_callback is not None and
                    hasattr(base_op.grad_callback.callback, base_op_grad_attr) and
                    base_op._ptr not in base_op_ptrs
                )

                if set_base_op_data_grad:
                    base_op.data = getattr(base_op.grad_callback.callback, base_op_grad_attr)
                    base_op_ptrs.append(base_op._ptr)
                else:
                    base_op.data = jnp.zeros(base_op.data.shape, dtype=base_op.dtype)

    return op_grad, 0.0, state_in_adj_bufs, params_grad


_operator_action.defvjp(_operator_action_fwd, _operator_action_bwd)



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

# Note: cuQuantum Python follows the cuQuantum SDK version, which is now
# switched to YY.MM and is different from individual libraries' (semantic)
# versioning scheme.

import os
import re
import sys
import shutil
import pathlib
import subprocess

from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from setuptools.command.bdist_wheel import bdist_wheel


try:
    cuda_path = os.environ['CUDA_PATH']
except KeyError as e:
    raise RuntimeError('CUDA is not found, please set $CUDA_PATH') from e


def check_cuda_version():
    try:
        # We cannot do a dlopen and call cudaRuntimeGetVersion, because it
        # requires GPUs. We also do not want to rely on the compiler utility
        # provided in distutils (deprecated) or setuptools, as this is a very
        # simple string parsing task.
        # TODO: switch to cudaRuntimeGetVersion once it's fixed (nvbugs 3624208)
        cuda_h = os.path.join(cuda_path, 'include', 'cuda.h')
        with open(cuda_h, 'r') as f:
            cuda_h = f.read()
        m = re.search('#define CUDA_VERSION ([0-9]*)', cuda_h)
        if m:
            ver = int(m.group(1))
        else:
            raise RuntimeError("cannot parse CUDA_VERSION")
    except:
        raise
    else:
        # 12020 -> "12.2"
        return str(ver // 1000) + '.' + str((ver % 100) // 10)


# We support CUDA 12/13 starting 25.09
cuda_ver = check_cuda_version()

if '12.0' <= cuda_ver < '13.0':
    cuda_major_ver = '12'
elif '13.0' <= cuda_ver < '14.0':
    cuda_major_ver = '13'
else:
    raise RuntimeError(f"Unsupported CUDA version: {cuda_ver}")


class CMakeExtension(Extension):
    def __init__(self, name: str, source_dir: str = ""):
        super().__init__(name, sources=[])
        self.source_dir = os.path.abspath(source_dir)

class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(["cmake", "--version"])
        except OSError:
            raise RuntimeError(
                "CMake must be installed to build the following extensions: "
                + ", ".join(e.name for e in self.extensions)
            )

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name)))
        cmake_args = []

        additional_cmake_args = os.environ.get("CMAKE_ARGS", "")
        if additional_cmake_args:
            cmake_args += additional_cmake_args.split()

        cfg = "Debug" if self.debug else "Release"
        build_args = ["--config", cfg]

        cmake_args += ["-DCMAKE_BUILD_TYPE=" + cfg]
        build_args += ["--", "-j2"]

        env = os.environ.copy()
        env["CXXFLAGS"] = '{} -DVERSION_INFO=\\"{}\\"'.format(
            env.get("CXXFLAGS", ""), self.distribution.get_version()
        )
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(
            ["cmake", ext.source_dir] + cmake_args, cwd=self.build_temp, env=env
        )
        subprocess.check_call(
            ["cmake", "--build", ".", "--verbose"] + build_args, cwd=self.build_temp
        )
        dso_dir = os.path.join(extdir, "cuquantum/lib")
        os.makedirs(dso_dir, exist_ok=True)
        for dso in pathlib.Path(self.build_temp).rglob("*.so"):
            shutil.copy2(dso, dso_dir)

class CustomBdistWheel(bdist_wheel):
    def finalize_options(self):
        super().finalize_options()
        self.root_is_pure = False

description = "NVIDIA cuQuantum Python JAX"
# Use README for the project long description
with open(os.path.join(os.path.dirname(__file__), "README.md"), encoding="utf-8") as f:
    long_description = f.read()

__version__ = "0.0.4"

classifiers=[
    "Development Status :: 5 - Production/Stable",
    "Operating System :: POSIX :: Linux",
    "Topic :: Education",
    "Topic :: Scientific/Engineering",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Programming Language :: Python :: 3.13",
    "Programming Language :: Python :: Implementation :: CPython",
    "Environment :: GPU :: NVIDIA CUDA"
]

install_requires = ['pybind11']
if cuda_major_ver == '12':
    classifiers.append("Environment :: GPU :: NVIDIA CUDA :: 12")
    install_requires.append("cuquantum-python-cu12~=26.1.0")
    install_requires.append("jax[cuda12-local]>=0.5,<0.7")
else:
    classifiers.append("Environment :: GPU :: NVIDIA CUDA :: 13")
    install_requires.append("cuquantum-python-cu13~=26.1.0")
    install_requires.append("jax[cuda13-local]>=0.8,<0.9")

setup(
    name="cuquantum-python-jax",
    version=__version__,
    url="https://developer.nvidia.com/cuquantum-sdk",
    author="NVIDIA Corporation",
    author_email="cuquantum-python@nvidia.com",
    python_requires=">=3.11.0",
    install_requires=install_requires,
    license="BSD-3-Clause",
    license_files = ('LICENSE',),
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=classifiers,
    ext_modules=[
        CMakeExtension("cppsrc", source_dir="cuquantum/densitymat/jax/cppsrc"),
    ],
    cmdclass=dict(
        build_ext=CMakeBuild,
        bdist_wheel=CustomBdistWheel
    ),
    packages=[
        "cuquantum.densitymat.jax",
        "cuquantum.densitymat.jax.pysrc"
    ],
    package_data={
        "cuquantum.densitymat.jax": ["**/*.so"]
    },
    exclude_package_data={
        "cuquantum.densitymat.jax": ["cppsrc/*"],
        "": ["tests/*"]
    },
    include_package_data=True,
    zip_safe=False
)



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Operator class in cuDensityMat.
"""

import ctypes
import logging
from typing import List, Tuple, Sequence

import jax
import jax.numpy as jnp

from cuquantum.bindings import cudensitymat as cudm

from .operator_term import OperatorTerm


@jax.tree_util.register_pytree_node_class
class Operator:
    """
    PyTree class for cuDensityMat's operator.
    """

    logger = logging.getLogger("cudensitymat-jax.Operator")

    def __init__(self, dims: Sequence[int]) -> None:
        """
        Initialize an Operator object.

        Args:
            dims: Hilbert space dimensions.
        """
        self.dims: Tuple[int, ...] = tuple(dims)

        self.op_terms: List[OperatorTerm] = []
        self.duals: List[bool] = []

        self.coeffs: List[float | complex] = []
        self.static_coeffs: List[jax.Array | None] = []
        self.total_coeffs: List[jax.ShapeDtypeStruct | None] = []
        self.coeff_callbacks: List[cudm.WrappedScalarCallback | None] = []
        self.coeff_grad_callbacks: List[cudm.WrappedScalarGradientCallback | None] = []

        self.batch_size: int = 1
        self.dtype: jnp.dtype | None = None

        self.static_coeffs_ptrs: List[int] = []
        self.total_coeffs_ptrs: List[int] = []
        self._static_coeffs_ptr_objs: List[ctypes.c_short] = []
        self._total_coeffs_ptr_objs: List[ctypes.c_short] = []

        self._batch_sizes: List[int] = []
        self._ptr: int | None = None

        self.op_terms_ids: List[int] = []

    def copy(self) -> "Operator":
        """
        Copy the operator.
        """
        op = Operator(self.dims)
        for (
            op_term,
            dual,
            coeff,
            static_coeffs,
            total_coeffs,
            coeff_callback,
            coeff_grad_callback
        ) in zip(
            self.op_terms,
            self.duals,
            self.coeffs,
            self.static_coeffs,
            self.total_coeffs,
            self.coeff_callbacks,
            self.coeff_grad_callbacks
        ):
            op_term_ = op_term.copy()
            static_coeffs_ = jnp.copy(static_coeffs) if static_coeffs is not None else None
            total_coeffs_ = total_coeffs if total_coeffs is not None else None
            op.append(op_term_,
                      dual=dual,
                      coeff=coeff,
                      static_coeffs=static_coeffs_,
                      total_coeffs=total_coeffs_,
                      coeff_callback=coeff_callback,
                      coeff_grad_callback=coeff_grad_callback)
        
        # Preserve pointer information to maintain PyTree structure consistency.
        # This is critical for custom VJP: the backward output must have the same
        # PyTree structure as the forward input, including pointer values.
        op.static_coeffs_ptrs = self.static_coeffs_ptrs.copy()
        op.total_coeffs_ptrs = self.total_coeffs_ptrs.copy()
        op._ptr = self._ptr
        op.op_terms_ids = self.op_terms_ids

        return op

    def tree_flatten(self):
        """
        Flatten the operator PyTree.
        """
        children = (self.op_terms, self.static_coeffs, self.static_coeffs_ptrs)
        aux_data = (
            self.dims,
            self.batch_size,
            self.duals,
            self.coeffs,
            self.total_coeffs,
            self.total_coeffs_ptrs,
            self.coeff_callbacks,
            self.coeff_grad_callbacks,
            self.dtype,
            self._batch_sizes,
            self._ptr,
            self.op_terms_ids
        )
        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """
        Unflatten the operator PyTree.
        """
        op_terms, static_coeffs, static_coeffs_ptrs = children
        (
            dims,
            batch_size,
            duals,
            coeffs,
            total_coeffs,
            total_coeffs_ptrs,
            coeff_callbacks,
            coeff_grad_callbacks,
            dtype,
            batch_sizes,
            ptr,
            op_terms_ids
        ) = aux_data

        inst = cls(dims)

        inst.op_terms = op_terms
        inst.static_coeffs = static_coeffs
        inst.static_coeffs_ptrs = static_coeffs_ptrs
        inst.total_coeffs = total_coeffs
        inst.total_coeffs_ptrs = total_coeffs_ptrs

        inst.batch_size = batch_size
        inst.duals = duals
        inst.coeffs = coeffs
        inst.coeff_callbacks = coeff_callbacks
        inst.coeff_grad_callbacks = coeff_grad_callbacks
        inst.dtype = dtype
        inst._batch_sizes = batch_sizes
        inst._ptr = ptr
        inst.op_terms_ids = op_terms_ids
        return inst

    def _check_dtype(self, op_term: OperatorTerm):
        """
        Check if all operator terms have the same data type.
        """
        if op_term.dtype is not None:  # for empty operator term, skip the check.
            if self.dtype is None:
                # If the data type is not set, set it to the data type of the first operator term.
                self.dtype = op_term.dtype
            else:
                # If the data type is set, check if the operator term has the same data type as the operator.
                if op_term.dtype != self.dtype:
                    raise ValueError("All operator terms must have the same data type.")

    def _check_and_append_batch_size(self,
                                     op_term: OperatorTerm,
                                     static_coeffs: jax.Array | None,
                                     total_coeffs: jax.Array | None
                                     ) -> None:
        """
        Check if all operator terms have batch size 1 or N.
        """
        # Extract the batch sizes of the static and total coefficients.
        if static_coeffs is not None and total_coeffs is None:
            coeffs_batch_size = static_coeffs.shape[0]
        elif static_coeffs is None and total_coeffs is not None:
            coeffs_batch_size = total_coeffs.shape[0]
        elif static_coeffs is not None and total_coeffs is not None:
            if static_coeffs.shape[0] != total_coeffs.shape[0]:
                raise ValueError("Static and total coefficients must have the same batch size if both are provided.")
            coeffs_batch_size = static_coeffs.shape[0]
        else:
            coeffs_batch_size = 1
        self._batch_sizes.append(coeffs_batch_size)

        # Possibly update the batch size of this operator and check consistency.
        batch_size = max(op_term.batch_size, coeffs_batch_size)
        if self.batch_size == 1:
            self.batch_size = batch_size
        else:
            if batch_size not in (1, self.batch_size):
                raise ValueError("Batch size in this operator term does not match batch size of this operator.")

    def append(self,
               op_term: OperatorTerm,
               *,
               dual: bool = False,
               coeff: float | complex = 1.0,
               static_coeffs: jax.Array | None = None,
               total_coeffs: jax.ShapeDtypeStruct | None = None,
               coeff_callback: cudm.WrappedScalarCallback | None = None,
               coeff_grad_callback: cudm.WrappedScalarGradientCallback | None = None
               ) -> None:
        """
        Append an operator term to an operator.

        Args:
            op_term: Operator term to be appended.
            dual: Duality of the operator term.
            coeff: Coefficient of the operator term.
            static_coeffs: Static batched coefficients of the operator term.
            total_coeffs: Total batched coefficients of the operator term.
            coeff_callback: Forward callback for the coefficients.
            coeff_grad_callback: Gradient callback for the coefficient.
        """
        if self._ptr is not None:
            raise RuntimeError("Cannot modify operator after it has been used in an operator action.")

        self.op_terms_ids.append(id(op_term))
    
        # Only require total_coeffs when using batched coefficients (static_coeffs provided)
        if coeff_callback is not None and static_coeffs is not None:
            if not isinstance(total_coeffs, jax.ShapeDtypeStruct):
                raise RuntimeError("For dynamically constructed batched coefficients, "
                                   "total_coeffs must be a jax.ShapeDtypeStruct.")

        # Check if the operator term has the same data type as the operator.
        self._check_dtype(op_term)
        self._check_and_append_batch_size(op_term, static_coeffs, total_coeffs)

        # Populate inst attributes.
        self.op_terms.append(op_term)
        self.duals.append(dual)

        self.coeffs.append(coeff)
        self.static_coeffs.append(static_coeffs)
        self.total_coeffs.append(total_coeffs)
        self.static_coeffs_ptrs.append(None)
        self.total_coeffs_ptrs.append(None)
        self.coeff_callbacks.append(coeff_callback)
        self.coeff_grad_callbacks.append(coeff_grad_callback)

    def __getitem__(self, index: int) -> OperatorTerm:
        """
        Get an operator term from the operator.
        """
        return self.op_terms[index]

    def _create(self, handle):
        """
        Create opaque handle to the operator.
        """
        # Create dependent operator terms.
        first_indices = {}
        for idx, val in enumerate(self.op_terms_ids):
            if val not in first_indices:
                first_indices[val] = idx
        mapping = {i: first_indices[val] for i, val in enumerate(self.op_terms_ids)}
        for index in set(mapping.values()):
            self.op_terms[index]._create(handle)

        # Create the current operator.
        if self._ptr is None:
            self._ptr = cudm.create_operator(
                handle,
                len(self.dims),
                self.dims
            )
            self.logger.debug(f"Created operator at {hex(self._ptr)}")

            for i in range(len(self.op_terms)):
                if self._batch_sizes[i] == 1:
                    cudm.operator_append_term(
                        handle,
                        self._ptr,
                        self.op_terms[mapping[i]]._ptr,
                        self.duals[i],
                        self.coeffs[i],
                        self.coeff_callbacks[i],
                        self.coeff_grad_callbacks[i]
                    )
                else:
                    if self.static_coeffs[i] is not None:
                        # Create and store object for temporary static coefficients.
                        static_coeffs_ptr_obj = ctypes.c_short()
                        self._static_coeffs_ptr_objs.append(static_coeffs_ptr_obj)

                        # Extract pointer of the object for temporary static coefficients.
                        static_coeffs_ptr = ctypes.addressof(static_coeffs_ptr_obj) + 1
                        assert static_coeffs_ptr % 2 == 1, "Temporary static coefficients pointer must be an odd number."
                        self.static_coeffs_ptrs[i] = static_coeffs_ptr
                    else:
                        static_coeffs_ptr = 0

                    if self.total_coeffs[i] is not None:
                        # Create and store object for temporary total coefficients.
                        total_coeffs_ptr_obj = ctypes.c_short()
                        self._total_coeffs_ptr_objs.append(total_coeffs_ptr_obj)

                        # Extract pointer of the object for temporary total coefficients.
                        total_coeffs_ptr = ctypes.addressof(total_coeffs_ptr_obj) + 1
                        assert total_coeffs_ptr % 2 == 1, "Temporary total coefficients pointer must be an odd number."
                        self.total_coeffs_ptrs[i] = total_coeffs_ptr
                    else:
                        total_coeffs_ptr = 0

                    cudm.operator_append_term_batch(
                        handle,
                        self._ptr,
                        self.op_terms[mapping[i]]._ptr,
                        self.duals[i],
                        self._batch_sizes[i],
                        static_coeffs_ptr,
                        total_coeffs_ptr,
                        self.coeff_callbacks[i],
                        self.coeff_grad_callbacks[i],
                    )
            self.logger.debug(f"Appended operator terms to operator at {hex(self._ptr)}")
    
    def _destroy(self):
        """
        Destroy opaque handle to the operator.
        """
        self._static_coeffs_ptr_objs.clear()
        self._total_coeffs_ptr_objs.clear()

        if self._ptr is not None:
            # Destroy the current operator.
            cudm.destroy_operator(self._ptr)
            self.logger.debug(f"Destroyed operator at {hex(self._ptr)}")
            self._ptr = None



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

"""
Operator action primitive.
"""

import logging
from typing import List, Tuple

import jax
import jax.numpy as jnp
from jax.interpreters import mlir

from cuquantum.bindings import cudensitymat as cudm

from .base import BasePrimitive, register_primitive
from .context import CudensitymatContext
from .operator import Operator


logger = logging.getLogger("cudensitymat-jax.operator_action")


class OperatorActionPrimitive(BasePrimitive):
    """
    JAX primitive for operator action.
    """

    name = "operator_action"
    inner_multiple_results = True
    outer_multiple_results = True
    inner_primitive = None
    outer_primitive = None
    operator_ptr = None  # Store only pointer to avoid leaking JAX tracers

    logger = logging.getLogger(f"cudensitymat-jax.OperatorActionPrimitive")

    @staticmethod
    def abstract(t_aval, params_aval, *other_in_buf_avals, op_ptr, other_out_shape_dtypes):
        """
        Abstract evaluation of the inner primitive of operator action.
        """
        OperatorActionPrimitive.logger.info(f"Calling abstract evaluation of the inner primitive")

        # TODO: Other output buffers, to be moved elsewhere.
        other_out_buf_avals = [
            jax.core.ShapedArray(shape_dtype.shape, shape_dtype.dtype)
            for shape_dtype in other_out_shape_dtypes
        ]

        # Get context using the stored pointer
        op_act_ctx = CudensitymatContext._contexts[OperatorActionPrimitive.operator_ptr]

        # Create abstract arrays for the output state buffers.
        num_state_components = op_act_ctx.num_state_components

        state_out_buf_avals = [
            jax.core.ShapedArray(other_in_buf_avals[i].shape, other_in_buf_avals[i].dtype)
            for i in range(num_state_components)
        ]

        # Obtain workspace limit and stream from the device.
        workspace_limit = (
            op_act_ctx.device.memory_stats()['bytes_limit'] -
            op_act_ctx.device.memory_stats()['bytes_in_use']
        )
        stream = op_act_ctx.device.get_stream_for_external_ready_events()

        # Prepare operator action.
        cudm.operator_prepare_action(
            CudensitymatContext._handle,
            op_act_ctx._operator,
            op_act_ctx._state_in,
            op_act_ctx._state_out,
            op_act_ctx._compute_type,
            workspace_limit,
            CudensitymatContext._workspace_desc,
            stream)

        # Query the required buffer size for the workspace.
        required_buffer_size = cudm.workspace_get_memory_size(
            CudensitymatContext._handle,
            CudensitymatContext._workspace_desc,
            cudm.Memspace.DEVICE,
            cudm.WorkspaceKind.WORKSPACE_SCRATCH)

        if required_buffer_size > op_act_ctx._required_buffer_size:
            op_act_ctx._required_buffer_size = required_buffer_size

        # Create abstract workspace array.
        # NOTE: Memory buffers from cudaMalloc is automatically 256-aligned, which is not 
        # the case for JAX. 255 is added to the buffer size to ensure workspace is 256-aligned.
        workspace_aval = jax.core.ShapedArray((op_act_ctx._required_buffer_size + 255,), jnp.uint8)
        return workspace_aval, *state_out_buf_avals, *other_out_buf_avals

    @staticmethod
    def outer_abstract(*args, **kwargs):
        """
        Abstract evaluation of the outer primitive of operator action.
        """
        OperatorActionPrimitive.logger.info(f"Calling abstract evaluation of the outer primitive")
        _, *out = OperatorActionPrimitive.abstract(*args, **kwargs)
        return out

    @staticmethod
    def lowering(ctx, t, params, *other_in_bufs, op_ptr, other_out_shape_dtypes):
        """
        Lowering rule of the operator action primitive.
        """
        OperatorActionPrimitive.logger.info(f"Calling lowering rule")

        # Get context using the stored pointer
        op_act_ctx = CudensitymatContext._contexts[OperatorActionPrimitive.operator_ptr]

        # Revert indices in input and output states. Note the layout is specified as
        # minor-to-major axis order.
        operand_layouts = [None] * len(ctx.avals_in)
        for i in range(2, len(ctx.avals_in)):  # 2 is for skipping t and params
            # 0 (the batch dimension) is the most major axis in input buffers and also when passed
            # to the cuQuantum library. Other dimensions (Hilbert space modes) need to be reversed.
            layout = tuple(range(1, ctx.avals_in[i].ndim)) + (0,)
            operand_layouts[i] = layout

        result_layouts = [None] * len(ctx.avals_out)
        for i in range(1, len(ctx.avals_out)):  # 1 is for skipping workspace
            # 0 (the batch dimension) is the most major axis in input buffers and also when passed
            # to the cuQuantum library. Other dimensions (Hilbert space modes) need to be reversed.
            layout = tuple(range(1, ctx.avals_out[i].ndim)) + (0,)
            result_layouts[i] = layout

        # Lower to the XLA FFI handler.
        outputs = jax.ffi.ffi_lowering(
            OperatorActionPrimitive.name,
            operand_layouts=operand_layouts,
            result_layouts=result_layouts
        )(
            ctx,
            t,
            params,
            *other_in_bufs,
            other_in_types=mlir.dense_int_elements(op_act_ctx.other_in_types),
            other_in_ptrs=mlir.dense_int_elements(op_act_ctx.other_in_ptrs),
            other_out_types=mlir.dense_int_elements(op_act_ctx.other_out_types),
            other_out_ptrs=mlir.dense_int_elements(op_act_ctx.other_out_ptrs),
            batch_size=op_act_ctx.batch_size,
            num_state_components=op_act_ctx.num_state_components,
            handle=CudensitymatContext._handle,
            operator=op_ptr,
            state_in=op_act_ctx._state_in,
            state_out=op_act_ctx._state_out
        )
        return outputs

    @staticmethod
    def impl(t, params, *other_in_bufs, op_ptr, other_out_shape_dtypes):
        """
        Primal evaluation of the operator action primitive.
        """
        OperatorActionPrimitive.logger.info(f"Calling primal evaluation")

        assert OperatorActionPrimitive.inner_primitive is not None
        _, *out = OperatorActionPrimitive.inner_primitive.bind(
            t,
            params,
            *other_in_bufs,
            op_ptr=op_ptr,
            other_out_shape_dtypes=tuple(other_out_shape_dtypes)
        )
        return out

    @staticmethod
    def batcher(batched_args, batch_dims, **kwargs):
        """
        Batching rule of the operator action primitive.
        """
        OperatorActionPrimitive.logger.info(f"Calling batcher")

        out = OperatorActionPrimitive.outer_primitive.bind(*batched_args, **kwargs)
        out_batched_dims = (0,) * len(out)
        return out, out_batched_dims


register_primitive(OperatorActionPrimitive)


def operator_action_prim(op: Operator,
                         t: float,
                         state_in_bufs: Tuple[jax.Array, ...],
                         params: jax.Array
                         ) -> List[jax.Array]:
    """
    Function wrapper around OperatorActionPrimitive.
    """
    logger.info(f"Calling operator_action_prim")

    # Extract unique batched coefficient buffers for operator terms and operator products.
    op_act_ctx = CudensitymatContext.get_context(op)

    other_in_bufs = []

    # Extract temporary batched coefficient buffers for operator terms.
    for i in op_act_ctx.op_term_coeffs_indices:
        other_in_bufs.append(op.static_coeffs[i])

    # Extract temporary batched coefficient buffers for operator products.
    static_coeffs = [x for op_term in op.op_terms for x in op_term.static_coeffs]
    for i in op_act_ctx.op_prod_coeffs_indices:
        other_in_bufs.append(static_coeffs[i])

    # Assign certain static attributes to the operator action context.
    base_ops = [base_op for op_term in op.op_terms for op_prod in op_term.op_prods for base_op in op_prod]
    for base_op in base_ops:
        if not base_op._to_be_allocated and base_op._ptr is not None:
            other_in_bufs.append(base_op.data)

    out = OperatorActionPrimitive.outer_primitive.bind(
        t,
        params,
        *state_in_bufs,
        *other_in_bufs,
        op_ptr=op._ptr,
        other_out_shape_dtypes=tuple(op_act_ctx.other_out_shape_dtypes)
    )

    state_out_bufs = out[:op_act_ctx.num_state_components]
    return state_out_bufs


class OperatorActionBackwardDiffPrimitive(BasePrimitive):
    """
    JAX primitive for operator action backward differentiation.
    """

    name = "operator_action_backward_diff"
    inner_multiple_results = True
    outer_multiple_results = True
    inner_primitive = None
    outer_primitive = None
    operator_ptr = None  # Store only pointer to avoid leaking JAX tracers

    logger = logging.getLogger("cudensitymat-jax.OperatorActionBackwardDiffPrimitive")

    @staticmethod
    def abstract(t_aval, params_aval, *other_in_buf_avals, op_ptr, other_out_shape_dtypes):
        """
        Abstract evaluation of the inner primitive of operator action backward differentiation.
        """
        OperatorActionBackwardDiffPrimitive.logger.info(f"Calling abstract evaluation of the inner primitive")

        other_out_buf_avals = [
            jax.core.ShapedArray(shape_dtype.shape, shape_dtype.dtype)
            for shape_dtype in other_out_shape_dtypes
        ]

        # Get context using the stored pointer
        op_act_ctx = CudensitymatContext._contexts[OperatorActionBackwardDiffPrimitive.operator_ptr]

        # Obtain number of state components.
        num_state_components = op_act_ctx.num_state_components

        # Create abstract arrays for the output quantities.
        params_grad_aval = jax.core.ShapedArray(params_aval.shape, params_aval.dtype)
        
        # Extract state input adjoint buffer shapes from state input buffers.
        state_in_adj_buf_avals = [
            jax.core.ShapedArray(other_in_buf_avals[i].shape, other_in_buf_avals[i].dtype)
            for i in range(num_state_components)
        ]

        # Obtain workspace limit and stream from the device.
        workspace_limit = (
            op_act_ctx.device.memory_stats()['bytes_limit'] -
            op_act_ctx.device.memory_stats()['bytes_in_use']
        )
        stream = op_act_ctx.device.get_stream_for_external_ready_events()

        # Prepare operator action backward differentiation.
        cudm.operator_prepare_action_backward_diff(
            CudensitymatContext._handle,
            op_act_ctx._operator,
            op_act_ctx._state_in,
            op_act_ctx._state_out_adj,
            op_act_ctx._compute_type,
            workspace_limit,
            CudensitymatContext._workspace_desc,
            stream)

        # Query the required buffer size for the workspace.
        required_buffer_size = cudm.workspace_get_memory_size(
            CudensitymatContext._handle,
            CudensitymatContext._workspace_desc,
            cudm.Memspace.DEVICE,
            cudm.WorkspaceKind.WORKSPACE_SCRATCH)
        if required_buffer_size > op_act_ctx._required_buffer_size:
            op_act_ctx._required_buffer_size = required_buffer_size

        # Create abstract workspace array.
        # NOTE: Memory buffers from cudaMalloc is automatically 256-aligned, which is not 
        # the case for JAX. 255 is added to the buffer size to ensure workspace is 256-aligned.
        workspace_aval = jax.core.ShapedArray((op_act_ctx._required_buffer_size + 255,), jnp.uint8)
        return workspace_aval, params_grad_aval, *state_in_adj_buf_avals, *other_out_buf_avals

    @staticmethod
    def outer_abstract(*args, **kwargs):
        """
        Abstract evaluation of the outer primitive of operator action backward differentiation.
        """
        OperatorActionBackwardDiffPrimitive.logger.info(f"Calling abstract evaluation of the outer primitive")
        _, *out = OperatorActionBackwardDiffPrimitive.abstract(*args, **kwargs)
        return out

    @staticmethod
    def lowering(ctx, t, params, *other_in_bufs, op_ptr, other_out_shape_dtypes):
        """
        Lowering rule of the operator action backward differentiation primitive.
        """
        OperatorActionBackwardDiffPrimitive.logger.info(f"Calling lowering rule")

        # Get context using the stored pointer
        op_act_ctx = CudensitymatContext._contexts[OperatorActionBackwardDiffPrimitive.operator_ptr]

        # Revert indices in input and output states. Note the layout is specified as
        # minor-to-major axis order.
        operand_layouts = [None] * len(ctx.avals_in)
        for i in range(2, len(ctx.avals_in)):  # 2 is for skipping t and params
            # 0 (the batch dimension) is the most major axis in input buffers and also when passed
            # to the cuQuantum library. Other dimensions (Hilbert space modes) need to be reversed.
            layout = tuple(range(1, ctx.avals_in[i].ndim)) + (0,)
            operand_layouts[i] = layout

        result_layouts = [None] * len(ctx.avals_out)
        for i in range(1, len(ctx.avals_out)):  # 1 is for skipping workspace and params_grad
            # 0 (the batch dimension) is the most major axis in output buffers and also when passed
            # to the cuQuantum library. Other dimensions (Hilbert space modes) need to be reversed.
            layout = tuple(range(1, ctx.avals_out[i].ndim)) + (0,)
            result_layouts[i] = layout

        # Lower to the XLA FFI handler.
        outputs = jax.ffi.ffi_lowering(
            OperatorActionBackwardDiffPrimitive.name,
            operand_layouts=operand_layouts,
            result_layouts=result_layouts
        )(
            ctx,
            t,
            params,
            *other_in_bufs,
            other_in_types=mlir.dense_int_elements(op_act_ctx.other_in_types),
            other_in_ptrs=mlir.dense_int_elements(op_act_ctx.other_in_ptrs),
            other_out_types=mlir.dense_int_elements(op_act_ctx.other_out_types),
            other_out_ptrs=mlir.dense_int_elements(op_act_ctx.other_out_ptrs),
            batch_size=op_act_ctx.batch_size,
            num_state_components=op_act_ctx.num_state_components,
            handle=CudensitymatContext._handle,
            operator=op_ptr,
            state_in=op_act_ctx._state_in,
            state_out_adj=op_act_ctx._state_out_adj,
            state_in_adj=op_act_ctx._state_in_adj
        )
        return outputs

    @staticmethod
    def impl(t, params, *other_in_bufs, op_ptr, other_out_shape_dtypes):
        """
        Primal evaluation of the operator action backward differentiation primitive.
        """
        OperatorActionBackwardDiffPrimitive.logger.info(f"Calling primal evaluation")
        assert OperatorActionBackwardDiffPrimitive.inner_primitive is not None
        _, *out = OperatorActionBackwardDiffPrimitive.inner_primitive.bind(
            t, params, *other_in_bufs, op_ptr=op_ptr, other_out_shape_dtypes=tuple(other_out_shape_dtypes))
        return out

    @staticmethod
    def batcher(batched_args, batch_dims, **kwargs):
        """
        Batching rule of the operator action backward differentiation primitive.
        """
        raise NotImplementedError("Batched operator action backward differentiation is not implemented.")


register_primitive(OperatorActionBackwardDiffPrimitive)


def operator_action_backward_diff_prim(op: Operator,
                                       t: float,
                                       state_in_bufs: Tuple[jax.Array, ...],
                                       state_out_adj_bufs: Tuple[jax.Array, ...],
                                       params: jax.Array
                                       ) -> Tuple[jax.Array, ...]:
    """
    Wrapper around the outer primitive of OperatorActionBackwardDiffPrimitive.
    """
    logger.info(f"Calling operator_action_backward_diff_prim")

    # Extract buffers from context.
    op_act_ctx = CudensitymatContext.get_context(op)

    other_in_bufs = []

    # Extract temporary batched coefficient buffers for operator terms.
    for i in op_act_ctx.op_term_coeffs_indices:
        other_in_bufs.append(op.static_coeffs[i])

    # Extract temporary batched coefficient buffers for operator products.
    static_coeffs = [x for op_term in op.op_terms for x in op_term.static_coeffs]
    for i in op_act_ctx.op_prod_coeffs_indices:
        other_in_bufs.append(static_coeffs[i])

    # Assign certain static attributes to the operator action context.
    base_ops = [base_op for op_term in op.op_terms for op_prod in op_term.op_prods for base_op in op_prod]
    for base_op in base_ops:
        if not base_op._to_be_allocated and base_op._ptr is not None:
            other_in_bufs.append(base_op.data)

    out = OperatorActionBackwardDiffPrimitive.outer_primitive.bind(
        t,
        params,
        *state_in_bufs,
        *state_out_adj_bufs,
        *other_in_bufs,
        op_ptr=op._ptr,
        other_out_shape_dtypes=tuple(op_act_ctx.other_out_shape_dtypes)
    )

    params_grad = out[0]
    state_in_adj_bufs = out[1:1 + op_act_ctx.num_state_components]
    return params_grad, *state_in_adj_bufs



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
#
# SPDX-License-Identifier: BSD-3-Clause

from .elementary_operator import ElementaryOperator
from .matrix_operator import MatrixOperator
from .operator_term import OperatorTerm
from .operator import Operator
